/***********************************************************************
 *
 * ivparse.c
 *
 * Read an Inventor model and make calls to specific functions named 
 * in graphics.h
 *
 * You shouldn't have to modify this file unless you wish to extend
 * the list of Inventor nodes handled by your program.
 *
 * AF971002
 *
 ***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <Inventor_c/SoDB.h>
#include <Inventor_c/SoInput.h>
#include <Inventor_c/SoPrimitiveVertex.h>
#include <Inventor_c/SbLinear.h>
#include <Inventor_c/actions/SoCallbackAction.h>
#include <Inventor_c/fields/SoSFVec3f.h>
#include <Inventor_c/fields/SoSFRotation.h>

#include <Inventor_c/nodes/SoSeparator.h>
#include <Inventor_c/nodes/SoPointLight.h>
#include <Inventor_c/nodes/SoMaterial.h>
#include <Inventor_c/nodes/SoTransform.h>
#include <Inventor_c/nodes/SoPerspectiveCamera.h>
#include <Inventor_c/nodes/SoSphere.h>
#include <Inventor_c/nodes/SoCone.h>
#include <Inventor_c/nodes/SoCube.h>
#include <Inventor_c/nodes/SoCylinder.h>
#include <Inventor_c/nodes/SoLabel.h>
#include <Inventor_c/nodes/SoCoordinate3.h>
#include <Inventor_c/nodes/SoFaceSet.h>
#include <Inventor_c/nodes/SoIndexedFaceSet.h>

#include "ivparse.h"
#include "graphics.h"


#define MAXTYPES 100

static SoType legalNodeTypes[MAXTYPES];
static int    numLegalNodeTypes;
static int    haveVisitedShape, haveVisitedCamera, haveVisitedLight;
static int    must_use_default_camera;
char  msgBuf[200];
char *badTypeMsg =
    "Encountered unknown node type '%s' in the Inventor file.\n\
     Known types are:\n\
     group, separator, perspective camera, point light, material, transform,\n\
     label, coord3, indexed face set, face set, cube, sphere, cone, cylinder";

void use_default_light(void), use_default_camera(void);

/***********************************************************************
 *
 * private functions
 *
 ***********************************************************************/


void vertexGetCoords(const SoPrimVtx *v, float *coords)
{
  const SbVec3f *point = SoPrimVtxGetPt(v);

  SbV3fGetXYZ(coords, *point);
}

void vertexGetNormal(const SoPrimVtx *v, float *normal)
{
  const SbVec3f *point = SoPrimVtxGetNorm(v);

  SbV3fGetXYZ(normal, *point);
}

void
triangleCallback(void *d, SoCBAct *a,
   const SoPrimVtx *vertex1,
   const SoPrimVtx *vertex2,
   const SoPrimVtx *vertex3)
{
  float v0[3], v1[3], v2[3];
  float n0[3], n1[3], n2[3];

  (void) (d, a);
  vertexGetCoords(vertex1, v0);
  vertexGetCoords(vertex2, v1);
  vertexGetCoords(vertex3, v2);

  vertexGetNormal(vertex1, n0);
  vertexGetNormal(vertex2, n1);
  vertexGetNormal(vertex3, n2);

  gr_triangle(v0,v1,v2,n0,n1,n2);
}


SoCallbackActionResponse
separatorStartCallback(void *d, SoCBAct *a, const SoNode *node)
{
  gr_separator_start();

  /* 
   * If there is no camera specified, insert a default camera
   * just after the first separator (root) of the scene graph.
   */
  if (must_use_default_camera) 
    {
      use_default_camera();
      must_use_default_camera = 0;
    }

  (void) (d, a, node);
  return SO_CB_ACT_CONTINUE;
}


SoCallbackActionResponse
separatorEndCallback(void *d, SoCBAct *a, const SoNode *node)
{
  gr_separator_end();

  (void) (d, a, node);
  return SO_CB_ACT_CONTINUE;
}

SoCallbackActionResponse
cameraCallback(void *d, SoCBAct *a, const SoNode *node)
{
  float position[3], orientation[4], nearDistance, farDistance, heightAngle;
  float angle;
  SbVec3f axis;

  SoPerspectiveCamera *camera = (SoPerspectiveCamera *) node;

  const SbVec3f    *vec = SoSV3fGet(&camera->position);
  const SbRotation *rot = SoSRotGet(&camera->orientation);

  (void) (d, a);
  SbV3fGetXYZ(position, *vec);

  SbRotGetAxisAngle(rot, &axis, &angle);
  SbV3fGetXYZ(orientation, axis);
  orientation[3] = angle;

  nearDistance = SoSFloatGet(&camera->nearDistance);
  farDistance  = SoSFloatGet(&camera->farDistance);
  heightAngle  = SoSFloatGet(&camera->heightAngle);
  
  gr_camera(position, orientation, nearDistance, farDistance, heightAngle);

  if (!haveVisitedLight) use_default_light();

  return SO_CB_ACT_CONTINUE;
}

SoCallbackActionResponse
pointLightCallback(void *d, SoCBAct *a, const SoNode *node)
{
  float intensity, color[3], location[3];

  SoPointLight *light = (SoPointLight *) node;

  const SbVec3f *vec = SoSV3fGet(&light->location);
  const SbVec3f *col = (SbVec3f *) SoSColGet(&light->color);

  (void) (d, a);
  SbV3fGetXYZ(location, *vec);
  SbV3fGetXYZ(color,    *col);

  intensity = SoSFloatGet(&light->intensity);

  gr_point_light(intensity, color, location);

  return SO_CB_ACT_CONTINUE;
}

SoCallbackActionResponse
materialCallback(void *d, SoCBAct *a, const SoNode *node)
{
  float ambientColor[3], diffuseColor[3], specularColor[3], emissiveColor[3],
    shininess, transparency;

  SoMaterial *material = (SoMaterial *) node;

  const SbVec3f *ambient  = (SbVec3f *) SoMColGet(&material->ambientColor, 0);
  const SbVec3f *diffuse  = (SbVec3f *) SoMColGet(&material->diffuseColor, 0);
  const SbVec3f *specular = (SbVec3f *) SoMColGet(&material->specularColor, 0);
  const SbVec3f *emissive = (SbVec3f *) SoMColGet(&material->emissiveColor, 0);

  (void) (d, a);
  SbV3fGetXYZ(ambientColor,  *ambient);
  SbV3fGetXYZ(diffuseColor,  *diffuse);
  SbV3fGetXYZ(specularColor, *specular);
  SbV3fGetXYZ(emissiveColor, *emissive);

  shininess = SoMFloatGet(&material->shininess,0);
  transparency = SoMFloatGet(&material->transparency,0);

  gr_material(ambientColor, diffuseColor, specularColor, emissiveColor,
	     shininess, transparency);

  return SO_CB_ACT_CONTINUE;
}

void
decomposeTransform(SoTransform *transform)
{
  float   x,y,z,t;
  SbVec3f axis;

  const SbVec3f    *trans = SoSV3fGet(&transform->translation);
  const SbRotation *rot   = SoSRotGet(&transform->rotation);
  const SbVec3f    *scale = SoSV3fGet(&transform->scaleFactor);

  SbV3fGetX_Y_Z(x, y, z, *trans);

  gr_translation(x,y,z);

  SbRotGetAxisAngle(rot, &axis, &t);
  SbV3fGetX_Y_Z(x, y, z, axis);

  gr_rotation(x,y,z,t);

  SbV3fGetX_Y_Z(x, y, z, *scale);

  gr_scale(x,y,z);
}

SoCallbackActionResponse
transformCallback(void *d, SoCBAct *a, const SoNode *node)
{
  SbVec3f origin;

  SoTransform *transform = (SoTransform *) node;

  (void) (d, a);
  SbV3fSetX_Y_Z(origin, (float) 0.0, (float) 0.0, (float) 0.0);

  SoXfRecenter(transform, &origin);

  decomposeTransform(transform);

  return SO_CB_ACT_CONTINUE;
}

SoCallbackActionResponse
labelCallback(void *d, SoCBAct *a, const SoNode *node)
{
  char *string;
  SoLabel *label = (SoLabel *) node;

  const SbName *name = SoSNameGet(&label->label);

  (void) (d, a);
  string = (char *) SbNameGetStr(name);

  gr_label(string);

  return SO_CB_ACT_CONTINUE;
}

SoCallbackActionResponse
cubeCallback(void *d, SoCBAct *a, const SoNode *node)
{
  float width,height,depth;

  SoCube *cube = (SoCube *) node;

  (void) (d, a);
  width  = SoSFloatGet(&cube->width);
  height = SoSFloatGet(&cube->height);
  depth  = SoSFloatGet(&cube->depth);

  gr_cube(width,height,depth);

  return SO_CB_ACT_CONTINUE;
}

SoCallbackActionResponse
sphereCallback(void *d, SoCBAct *a, const SoNode *node)
{
  float radius;

  SoSphere *sphere = (SoSphere *) node;

  (void) (d, a);
  radius = SoSFloatGet(&sphere->radius);

  gr_sphere(radius);

  return SO_CB_ACT_CONTINUE;
}

SoCallbackActionResponse
coneCallback(void *d, SoCBAct *a, const SoNode *node)
{
  float radius, height;

  SoCone *cone = (SoCone *) node;

  (void) (d, a);
  radius = SoSFloatGet(&cone->bottomRadius);
  height = SoSFloatGet(&cone->height);

  gr_cone(radius, height);

  return SO_CB_ACT_CONTINUE;
}

SoCallbackActionResponse
cylinderCallback(void *d, SoCBAct *a, const SoNode *node)
{
  float radius, height;

  SoCylinder *cyl = (SoCylinder *) node;

  (void) (d, a);
  radius = SoSFloatGet(&cyl->radius);
  height = SoSFloatGet(&cyl->height);

  gr_cylinder(radius, height);

  return SO_CB_ACT_CONTINUE;
}

void
traverseGraph(SoNode *root,
	      int cubeTriangles,    /* 0 for cube, 1 for triangles */
	      int sphereTriangles,  /* etc. */
	      int coneTriangles,
	      int cylinderTriangles)
{
   SoCBAct *myAction;

   myAction = SoCBActCreate();

   /* 
    * pre and post callbacks for separator 
    */
   SoCBActAddPreCB(myAction, SoSepGetClassTypeId(), 
		   separatorStartCallback, NULL);

   SoCBActAddPostCB(myAction, SoSepGetClassTypeId(), 
		   separatorEndCallback, NULL);

   /* 
    * pre callbacks for camera, light, material, xform, label
    */
   SoCBActAddPreCB(myAction, SoPerspCamGetClassTypeId(), 
            cameraCallback, NULL);

   SoCBActAddPreCB(myAction, SoPtLtGetClassTypeId(), 
            pointLightCallback, NULL);

   SoCBActAddPreCB(myAction, SoMtlGetClassTypeId(), 
            materialCallback, NULL);

   SoCBActAddPreCB(myAction, SoXfGetClassTypeId(), 
            transformCallback, NULL);

   SoCBActAddPreCB(myAction, SoLabelGetClassTypeId(), 
            labelCallback, NULL);
   
   /*
    * triangle callback for indexed face set and indexed face set
    */
   SoCBActAddTriCB(myAction, SoIndFSetGetClassTypeId(), 
		   triangleCallback, NULL);

   SoCBActAddTriCB(myAction, SoFSetGetClassTypeId(), 
		   triangleCallback, NULL);

   /* 
    * triangle or regular callbacks for sphere, cone, cube, and cylinder
    */
   if (cubeTriangles)
     SoCBActAddTriCB(myAction, SoCubeGetClassTypeId(), triangleCallback, NULL);
   else
     SoCBActAddPreCB(myAction, SoCubeGetClassTypeId(), cubeCallback, NULL);

   if (sphereTriangles)
     SoCBActAddTriCB(myAction, SoSphGetClassTypeId(), triangleCallback, NULL);
   else
     SoCBActAddPreCB(myAction, SoSphGetClassTypeId(), sphereCallback, NULL);

   if (coneTriangles)
     SoCBActAddTriCB(myAction, SoConeGetClassTypeId(), triangleCallback, NULL);
   else
     SoCBActAddPreCB(myAction, SoConeGetClassTypeId(), coneCallback, NULL);

   if (cylinderTriangles)
     SoCBActAddTriCB(myAction, SoCylGetClassTypeId(), triangleCallback, NULL);
   else
     SoCBActAddPreCB(myAction, SoCylGetClassTypeId(), cylinderCallback, NULL);

   must_use_default_camera = (!haveVisitedCamera);

   /* now do the traversal */
   SoCBActApply(myAction, root);
   SoCBActDelete(myAction);
}

void
die(char *msg)
{
  fprintf(stderr, "%s\n", msg);
  exit(-1);
}

void
setupLegalNodeTypes(void)
{
  numLegalNodeTypes = 0;
  haveVisitedCamera = 0;
  haveVisitedLight  = 0;
  haveVisitedShape  = 0;
  /* 
   * There should be less than MAXTYPES types added to this list.
   * Right now, legal types are:
   *  group, separator, perspective camera, point light, material, transform,
   *  label, coord3, indexed face set, face set, cube, sphere, cone, cylinder
   */
  legalNodeTypes[numLegalNodeTypes++] = SoGroupGetClassTypeId();
  legalNodeTypes[numLegalNodeTypes++] = SoSepGetClassTypeId();
  legalNodeTypes[numLegalNodeTypes++] = SoPerspCamGetClassTypeId();
  legalNodeTypes[numLegalNodeTypes++] = SoPtLtGetClassTypeId();
  legalNodeTypes[numLegalNodeTypes++] = SoMtlGetClassTypeId();
  legalNodeTypes[numLegalNodeTypes++] = SoXfGetClassTypeId();
  legalNodeTypes[numLegalNodeTypes++] = SoLabelGetClassTypeId();
  legalNodeTypes[numLegalNodeTypes++] = SoCoord3GetClassTypeId();
  legalNodeTypes[numLegalNodeTypes++] = SoIndFSetGetClassTypeId();
  legalNodeTypes[numLegalNodeTypes++] = SoFSetGetClassTypeId();
  legalNodeTypes[numLegalNodeTypes++] = SoCubeGetClassTypeId();
  legalNodeTypes[numLegalNodeTypes++] = SoSphGetClassTypeId();
  legalNodeTypes[numLegalNodeTypes++] = SoConeGetClassTypeId();
  legalNodeTypes[numLegalNodeTypes++] = SoCylGetClassTypeId();
}

char *nodeTypeName(SoType nodeType)
{
  SbName typeName;
  const char  *strName;

  typeName = SoTypeGetName(&nodeType);
  strName  = SbNameGetStr(&typeName);

  return (char *) strName;
}

int
nodeTypeIsGroup(SoType nodeType)
{
  SoType groupType = SoGroupGetClassTypeId();

  return (SoTypeIsEq(&nodeType, groupType) ||
	  SoTypeIsDerivedFrom(&nodeType, groupType));
}


void
verifyNode(SoNode *node)
{
  int i;
  SoType nodeType = SoNodeGetTypeId(node);

  if (SoTypeIsEq(&nodeType, SoPerspCamGetClassTypeId()))
    {
      haveVisitedCamera = 1;

      if (haveVisitedShape)
	{
	  die("Please specify camera before other leaf nodes in iv file.");
	}
    }
  else if (!nodeTypeIsGroup(nodeType))
      {
	haveVisitedShape = 1;

	if (SoTypeIsEq(&nodeType, SoPtLtGetClassTypeId()))
	  {
	    haveVisitedLight = 1;
	  }
      }

  for (i = 0; i < numLegalNodeTypes; i++)
    {
      if (SoTypeIsEq(&nodeType, legalNodeTypes[i])) return;
    }

  sprintf(msgBuf, badTypeMsg, nodeTypeName(nodeType));
  die(msgBuf);
}

void
verifyGroup(SoNode *graph)
{
  int children,i;

  SoType nodeType  = SoNodeGetTypeId(graph);

  verifyNode(graph);

  if (nodeTypeIsGroup(nodeType))
    {
      children = SoGroupGetNumChildren((SoGroup *) graph);
      
      for (i = 0; i < children; i++)
	{
	  verifyGroup(SoGroupGetChild((SoGroup *) graph,i));
	}
    }
}

void
verifyGraph(SoNode *graph)
{
  setupLegalNodeTypes();

  verifyGroup(graph);

  if (!haveVisitedCamera)
    {
      fprintf(stderr, 
        "warning: No camera is specified in inventor file. Using default.\n");
    }
  if (!haveVisitedLight)
    {
      fprintf(stderr, 
        "warning: No light is specified in inventor file. Using default.\n");
    }
}

void
use_default_camera(void)
{
  float position[3], orientation[4], nearDistance, farDistance, heightAngle;

  position[0] = (float) 2.0;
  position[1] = (float) 0.5;
  position[2] = (float) 2.0;

  orientation[0] = (float) 0.0;
  orientation[1] = (float) 1.0;
  orientation[2] = (float) 0.0;
  orientation[3] = (float) 0.786;

  nearDistance = (float) 0.004;
  farDistance  = (float) 3.5;
  heightAngle  = (float) 0.7854;

  gr_camera(position, orientation, nearDistance, farDistance, heightAngle);

  if (!haveVisitedLight) use_default_light();
}

void
use_default_light(void)
{
  float intensity, color[3], location[3];

  intensity = (float) 1.0;

  color[0]  = (float) 0.5;
  color[1]  = (float) 0.5;
  color[2]  = (float) 0.5;

  location[0] = (float) 10.0;
  location[1] = (float) 5.0;
  location[2] = (float) 8.0;

  gr_point_light(intensity, color, location);
}


/***********************************************************************
 *
 * exported
 *
 ***********************************************************************/

static SoSep *root = NULL;

void iv_parse(char *filename)
{
  SoIn  *mySceneInput;

  if (root != NULL)
    {
      SoSepUnref(root);		/* Free memory for old scene graph */
    }
  else
    {
      SoDBInit();		/* Initialize Inventor */
    }

  /* Read object data from a file */
  mySceneInput = SoInCreate();
  if (!SoInOpenFile(mySceneInput, filename, FALSE) ||
      !(root = SoDBReadAll(mySceneInput)))
    {
      fprintf(stderr, "Failed to read scene in inventor file %s\n", filename);
      exit (1);
    }
  SoSepRef(root);
  SoInCloseFile(mySceneInput);
  SoInDelete(mySceneInput);

  verifyGraph((SoNode *) root);
}

void iv_traverse(int cubeTriangles,    /* 0 for cube, 1 for triangles */
		int sphereTriangles,  /* etc. */
		int coneTriangles,
		int cylinderTriangles)
{
  if (!root) return;

  gr_traversal_start();

  traverseGraph((SoNode *) root,
		cubeTriangles, sphereTriangles, 
		coneTriangles, cylinderTriangles);

  gr_traversal_end();
}
